"use strict";
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlContainerManager = void 0;
const langium_1 = require("langium");
const ast_1 = require("./generated/ast");
class SqlContainerManager {
    constructor(services) {
        this.children = new langium_1.MultiMap();
        this.langiumDocuments = services.workspace.LangiumDocuments;
        services.workspace.DocumentBuilder.onBuildPhase(langium_1.DocumentState.ComputedScopes, (_, cancelToken) => this.rebuild(cancelToken));
    }
    getChildren(definition) {
        return this.children.get(definition);
    }
    putChild(definition, child) {
        this.children.add(definition, child);
    }
    async rebuild(cancelToken) {
        this.children.clear();
        const documents = this.langiumDocuments.all;
        const sink = {
            schemas: [],
            tables: [],
            functions: []
        };
        for (const document of documents) {
            const value = document.parseResult.value;
            if ((0, ast_1.isSqlFile)(value)) {
                await this.collectDocument(value, sink, cancelToken);
            }
        }
        await this.collectElements(sink.schemas, cancelToken);
        await this.collectElements(sink.tables, cancelToken);
        await this.collectElements(sink.functions, cancelToken);
    }
    async collectElements(elements, cancelToken) {
        var _a;
        for (const element of elements) {
            await (0, langium_1.interruptAndCheck)(cancelToken);
            const reference = element.reference;
            if (reference === null || reference === void 0 ? void 0 : reference.previous) {
                const previousElement = (_a = reference.previous.element) === null || _a === void 0 ? void 0 : _a.ref;
                if (previousElement) {
                    this.putChild(previousElement, element);
                }
            }
        }
    }
    async collectDocument(file, sink, cancelToken) {
        for (const statement of file.statements) {
            await (0, langium_1.interruptAndCheck)(cancelToken);
            if ((0, ast_1.isSchemaDefinition)(statement)) {
                sink.schemas.push(statement);
            }
            else if ((0, ast_1.isTableDefinition)(statement)) {
                sink.tables.push(statement);
            }
            else if ((0, ast_1.isFunctionDefinition)(statement)) {
                sink.functions.push(statement);
            }
        }
    }
}
exports.SqlContainerManager = SqlContainerManager;
//# sourceMappingURL=sql-container-manager.js.map