"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesExtend = exports.DialectTypes = exports.parseRequiredType = exports.isCompatibleWithDefinition = exports.toString = void 0;
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
const ast_1 = require("./generated/ast");
function toString(dataType) {
    const name = `${dataType.names.join(' ')}`;
    if (dataType.arguments.length > 0) {
        const args = `(${dataType.arguments.map(arg => `${arg.type}${arg.optional ? '?' : ''}`).join(', ')})`;
        return `${name}${args}`;
    }
    return name;
}
exports.toString = toString;
function isCompatibleWithDefinition(ast, required) {
    if (ast.dataTypeNames.length !== required.names.length) {
        return false;
    }
    for (let i = 0; i < ast.dataTypeNames.length; i++) {
        if (ast.dataTypeNames[i].toUpperCase() !== required.names[i].toUpperCase()) {
            return false;
        }
    }
    if (ast.arguments.length > required.arguments.length) {
        return false;
    }
    for (let i = 0; i < ast.arguments.length; i++) {
        const arg = ast.arguments[i];
        const requiredArg = required.arguments[i];
        if ((0, ast_1.isNumberLiteral)(arg)) {
            switch (requiredArg.type) {
                case 'integer':
                    if (!Number.isInteger(arg.value)) {
                        return false;
                    }
                    break;
                case 'real':
                    break;
                case 'size':
                    if (!Number.isInteger(arg.value) || arg.value < 0) {
                        return false;
                    }
                    break;
                default:
                    return false;
            }
        }
        if ((0, ast_1.isStringLiteral)(arg) && requiredArg.type !== 'string') {
            return false;
        }
    }
    for (let i = ast.arguments.length; i < required.arguments.length; i++) {
        if (!required.arguments[i].optional) {
            return false;
        }
    }
    return true;
}
exports.isCompatibleWithDefinition = isCompatibleWithDefinition;
function parseRequiredType(str) {
    const namesExpression = /^(\w+)(\s+(\w+))*\s*(\([^\)]+\))?$/;
    const namesMatch = str.match(namesExpression);
    if (!namesMatch) {
        throw new Error(`Invalid type string: ${str}`);
    }
    const names = [];
    for (let index = 1; index < namesMatch.length; index += 2) {
        const name = namesMatch[index];
        name && names.push(name);
    }
    const args = [];
    const argsString = namesMatch[namesMatch.length - 1];
    if (argsString) {
        const trimmedArgsString = argsString.substring(1, argsString.length - 1).split(',').map(s => s.trim());
        for (const arg of trimmedArgsString) {
            const match = arg.match(/^(string|real|integer|size)(\?)?$/);
            if (match) {
                args.push({
                    type: match[1],
                    optional: match[2] === '?'
                });
            }
            else {
                throw new Error(`Invalid argument type: ${arg}`);
            }
        }
    }
    return {
        names,
        arguments: args
    };
}
exports.parseRequiredType = parseRequiredType;
class DialectTypes {
    constructor(types) {
        this.types = Object.entries(types).reduce((acc, [key, value]) => {
            acc[key] = value.map(parseRequiredType);
            return acc;
        }, {});
    }
    isStringDataType(dataType) {
        return this.types.strings.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isIntegerDataType(dataType) {
        return this.types.integers.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isRealDataType(dataType) {
        return this.types.reals.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isDateTimeDataType(dataType) {
        return this.types.dateTimes.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isBooleanDataType(dataType) {
        return this.types.booleans.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isBlobDataType(dataType) {
        return this.types.blobs.some(t => isCompatibleWithDefinition(dataType, t));
    }
    isEnumDataType(dataType) {
        return this.types.enums.some(t => isCompatibleWithDefinition(dataType, t));
    }
    allTypes() {
        return [
            ...this.types.strings,
            ...this.types.integers,
            ...this.types.reals,
            ...this.types.dateTimes,
            ...this.types.booleans,
            ...this.types.blobs,
            ...this.types.enums
        ];
    }
}
exports.DialectTypes = DialectTypes;
function typesExtend(a, b) {
    return {
        strings: [...a.strings, ...b.strings],
        integers: [...a.integers, ...b.integers],
        reals: [...a.reals, ...b.reals],
        dateTimes: [...a.dateTimes, ...b.dateTimes],
        booleans: [...a.booleans, ...b.booleans],
        blobs: [...a.blobs, ...b.blobs],
        enums: [...a.enums, ...b.enums]
    };
}
exports.typesExtend = typesExtend;
//# sourceMappingURL=sql-data-types.js.map