"use strict";
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlScopeProvider = void 0;
const langium_1 = require("langium");
const ast_1 = require("./generated/ast");
const sql_type_utilities_1 = require("./sql-type-utilities");
class SqlScopeProvider extends langium_1.DefaultScopeProvider {
    constructor(services) {
        super(services);
        this.astNodeDescriptionProvider = services.workspace.AstNodeDescriptionProvider;
        this.astNodeLocator = services.workspace.AstNodeLocator;
        this.langiumDocuments = services.shared.workspace.LangiumDocuments;
        this.containerManager = services.shared.workspace.ContainerManager;
    }
    getScope(context) {
        var _a, _b, _c;
        const container = context.container;
        if ((0, ast_1.isGlobalReference)(container)) {
            const previous = container.previous;
            if (previous) {
                const element = (_a = previous.element) === null || _a === void 0 ? void 0 : _a.ref;
                if (element) {
                    return this.createScopeForNodes(this.containerManager.getChildren(element));
                }
                else {
                    return langium_1.EMPTY_SCOPE;
                }
            }
            else {
                // Use lexical scoping if no previous element has been assigned
                return super.getScope(context);
            }
        }
        else if ((0, ast_1.isPrimaryKeyDefinition)(container)) {
            const property = context.property;
            switch (property) {
                case 'primaryKeys':
                    const tableDef = (0, langium_1.getContainerOfType)(container, ast_1.isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(ast_1.isColumnDefinition));
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isIndexDefinition)(container)) {
            const property = context.property;
            switch (property) {
                case 'indexes':
                    const tableDef = (0, langium_1.getContainerOfType)(container, ast_1.isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(ast_1.isColumnDefinition));
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isKeyDefinition)(container)) {
            const property = context.property;
            switch (property) {
                case 'keys':
                    const tableDef = (0, langium_1.getContainerOfType)(container, ast_1.isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(ast_1.isColumnDefinition));
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isConstraintDefinition)(container)) {
            const property = context.property;
            switch (property) {
                case "from": {
                    const columns = (0, langium_1.getContainerOfType)(container, ast_1.isTableDefinition).columns.filter(ast_1.isColumnDefinition);
                    return this.streamColumnDefinitions(columns);
                }
                case 'to': {
                    const columns = (_b = (0, sql_type_utilities_1.getFromGlobalReference)(container.table, ast_1.isTableDefinition)) === null || _b === void 0 ? void 0 : _b.columns.filter(ast_1.isColumnDefinition);
                    return this.streamColumnDefinitions(columns !== null && columns !== void 0 ? columns : []);
                }
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isAllTable)(container)) {
            const property = context.property;
            switch (property) {
                case "variableName": {
                    const selectStatement = (0, langium_1.getContainerOfType)(container, ast_1.isSimpleSelectStatement);
                    //ATTENTION! Do not recursively traverse upwards, t.* only looks up t in the current select statement
                    return this.newCaseInsensitiveScope((0, langium_1.stream)(this.getTableVariablesForSelectStatement(selectStatement)));
                }
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isTableRelatedColumnExpression)(container)) {
            const property = context.property;
            switch (property) {
                case "variableName":
                    const selectStatement = (0, langium_1.getContainerOfType)(container, ast_1.isSimpleSelectStatement);
                    return this.getTableVariablesForSelectStatementRecursively(context, selectStatement);
                case "columnName":
                    const sourceItem = (_c = container.variableName) === null || _c === void 0 ? void 0 : _c.ref;
                    if (sourceItem) {
                        if ((0, ast_1.isTableSourceItem)(sourceItem)) {
                            const tableLike = (0, sql_type_utilities_1.getFromGlobalReference)(sourceItem.table, ast_1.isTableLike);
                            if (tableLike) {
                                if ((0, ast_1.isTableDefinition)(tableLike)) {
                                    return this.streamColumnDefinitions(tableLike.columns.filter(ast_1.isColumnDefinition));
                                }
                                else if ((0, ast_1.isCommonTableExpression)(tableLike)) {
                                    const candidates = (0, sql_type_utilities_1.getColumnCandidatesForSelectTableExpression)(tableLike.statement);
                                    return this.streamColumnDescriptors(candidates);
                                }
                                else {
                                    (0, langium_1.assertUnreachable)(tableLike);
                                }
                            }
                        }
                        else if ((0, ast_1.isSubQuerySourceItem)(sourceItem)) {
                            const selectStatement = sourceItem.subQuery;
                            const candidates = (0, sql_type_utilities_1.getColumnCandidatesForSelectTableExpression)(selectStatement);
                            return this.streamColumnDescriptors(candidates);
                        }
                        else {
                            (0, langium_1.assertUnreachable)(sourceItem);
                        }
                    }
                    break;
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isColumnNameExpression)(container)) {
            const property = context.property;
            switch (property) {
                case 'columnName':
                    const selectStatement = (0, langium_1.getContainerOfType)(container, ast_1.isSimpleSelectStatement);
                    const candidates = (0, sql_type_utilities_1.getColumnCandidatesForSimpleSelectStatement)(selectStatement);
                    return this.streamColumnDescriptors(candidates);
                default:
                    (0, langium_1.assertUnreachable)(property);
            }
        }
        else if ((0, ast_1.isOverClause)(container)) {
            const property = context.property;
            switch (property) {
                case 'windowName':
                    const selectStatement = (0, langium_1.getContainerOfType)(container, ast_1.isSimpleSelectStatement);
                    const nodes = (0, langium_1.streamAllContents)(selectStatement).filter(ast_1.isWindowSpec).toArray();
                    return this.packToDescriptions(nodes);
            }
        }
        else {
            (0, langium_1.assertUnreachable)(container);
        }
        return super.getScope(context);
    }
    getTableVariablesForSelectStatementRecursively(context, selectStatement) {
        let outerScope = undefined;
        if (selectStatement && (0, langium_1.hasContainerOfType)(selectStatement.$container, ast_1.isSimpleSelectStatement)) {
            const outerSelectStatement = (0, langium_1.getContainerOfType)(selectStatement.$container, ast_1.isSimpleSelectStatement);
            outerScope = this.getTableVariablesForSelectStatementRecursively(context, outerSelectStatement);
        }
        const descriptions = this.getTableVariablesForSelectStatement(selectStatement);
        return this.newCaseInsensitiveScope((0, langium_1.stream)(descriptions), outerScope);
    }
    getTableVariablesForSelectStatement(selectStatement) {
        if (selectStatement === null || selectStatement === void 0 ? void 0 : selectStatement.from) {
            const astDescriptions = [];
            for (const source of selectStatement.from.sources.list) {
                const items = [source.item].concat(source.joins.map((j) => j.nextItem));
                for (const item of items) {
                    if (item.name) {
                        astDescriptions.push(this.astNodeDescriptionProvider.createDescription(item, item.name));
                    }
                    else {
                        if ((0, ast_1.isTableSourceItem)(item)) {
                            const tableLike = (0, sql_type_utilities_1.getFromGlobalReference)(item.table, ast_1.isTableLike);
                            if (tableLike) {
                                const name = this.nameProvider.getName(tableLike);
                                if (name) {
                                    astDescriptions.push(this.astNodeDescriptionProvider.createDescription(item, name));
                                }
                            }
                        }
                        else if (!(0, ast_1.isSubQuerySourceItem)(item)) {
                            (0, langium_1.assertUnreachable)(item);
                        }
                    }
                }
            }
            return astDescriptions;
        }
        return [];
    }
    getGlobalScope(nodeType, _context) {
        return this.newCaseInsensitiveScope(this.indexManager.allElements(nodeType));
    }
    unpackFromDescriptions(scope) {
        return scope.getAllElements().toArray()
            .map(d => {
            const document = this.langiumDocuments.getOrCreateDocument(d.documentUri);
            return this.astNodeLocator.getAstNode(document.parseResult.value, d.path);
        });
    }
    packToDescriptions(nodes) {
        return this.newCaseInsensitiveScope((0, langium_1.stream)(nodes.map((c) => this.astNodeDescriptionProvider.createDescription(c, c.name))));
    }
    streamColumnDescriptors(columns) {
        return this.newCaseInsensitiveScope((0, langium_1.stream)(columns.filter(c => c.name).map(c => this.astNodeDescriptionProvider.createDescription(c.node, c.name))));
    }
    streamColumnDefinitions(columns) {
        return this.packToDescriptions(columns);
    }
    newCaseInsensitiveScope(stream, outerScope = undefined) {
        return new langium_1.StreamScope(stream, outerScope, { caseInsensitive: true });
    }
}
exports.SqlScopeProvider = SqlScopeProvider;
//# sourceMappingURL=sql-scope-provider.js.map