"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlTypeComputer = void 0;
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
const langium_1 = require("langium");
const ast_1 = require("./generated/ast");
const sql_type_conversion_1 = require("./sql-type-conversion");
const sql_type_descriptors_1 = require("./sql-type-descriptors");
const sql_type_operators_1 = require("./sql-type-operators");
const sql_type_utilities_1 = require("./sql-type-utilities");
class SqlTypeComputer {
    constructor(services) {
        this.dataTypes = services.dialect.dataTypes;
    }
    computeType(node) {
        if ((0, ast_1.isExpression)(node)) {
            return this.computeTypeOfExpression(node);
        }
        else if ((0, ast_1.isType)(node)) {
            return this.computeTypeOfDataType(node);
        }
        return undefined;
    }
    computeTypeOfExpression(node) {
        if ((0, ast_1.isCastExpression)(node)) {
            const source = this.computeType(node.expr);
            const target = this.computeTypeOfDataType(node.type);
            return source && target && (0, sql_type_conversion_1.canConvert)(source, target, 'explicit') ? target : undefined;
        }
        if ((0, ast_1.isNumberLiteral)(node)) {
            return this.computeTypeOfNumericLiteral(node.$cstNode.text);
        }
        if ((0, ast_1.isNullLiteral)(node)) {
            return sql_type_descriptors_1.Types.Null;
        }
        if ((0, ast_1.isHexStringLiteral)(node)) {
            return sql_type_descriptors_1.Types.Integer;
        }
        if ((0, ast_1.isTableRelatedColumnExpression)(node)) { //variable.columnName
            const varRef = node.variableName.ref;
            if (!varRef) {
                return undefined;
            }
            else if ((0, ast_1.isTableSourceItem)(varRef)) { //tableVariable.columnName
                const ref = node.columnName.ref;
                if (!(0, ast_1.isColumnDefinition)(ref)) {
                    return undefined;
                }
                return this.computeType(ref.dataType);
            }
            else if ((0, ast_1.isSubQuerySourceItem)(varRef)) { //subqueryVariable.selectItemName
                const ref = node.columnName.ref;
                if (!(0, ast_1.isExpressionQuery)(ref)) {
                    return undefined;
                }
                return this.computeType(ref.expr);
            }
            else {
                (0, langium_1.assertUnreachable)(varRef);
                return undefined;
            }
        }
        if ((0, ast_1.isParenthesisOrListExpression)(node)) {
            const firstType = this.computeType(node.items[0]);
            //ONLY the IN operator is allowed to look up a list!
            if (firstType && node.$container.$type === ast_1.NegatableExpression && node.$container.operator === 'IN') {
                return sql_type_descriptors_1.Types.ArrayOf(firstType);
            }
            return firstType;
        }
        if ((0, ast_1.isUnaryExpression)(node)) {
            const operandType = this.computeType(node.value);
            return operandType != null
                ? this.computeTypeOfUnaryOperation(node.operator, operandType)
                : undefined;
        }
        if ((0, ast_1.isStringLiteral)(node)) {
            return sql_type_descriptors_1.Types.Char();
        }
        if ((0, ast_1.isBooleanLiteral)(node)) {
            return sql_type_descriptors_1.Types.Boolean;
        }
        if ((0, ast_1.isColumnNameExpression)(node)) {
            const ref = node.columnName.ref;
            if (!ref) {
                return undefined;
            }
            else if ((0, ast_1.isExpressionQuery)(ref)) {
                return this.computeType(ref.expr);
            }
            else if ((0, ast_1.isColumnDefinition)(ref)) {
                return this.computeType(ref.dataType);
            }
            else if ((0, ast_1.isCteColumnName)(ref)) {
                throw new Error('TODO');
            }
            else {
                (0, langium_1.assertUnreachable)(ref);
            }
        }
        if ((0, ast_1.isFunctionCall)(node)) {
            const functionLike = (0, sql_type_utilities_1.getFromGlobalReference)(node.function, ast_1.isFunctionDefinition);
            if (functionLike) {
                return this.computeTypeOfDataType(functionLike.returnType);
            }
            else {
                return undefined;
            }
        }
        if ((0, ast_1.isBinaryExpression)(node) || (0, ast_1.isNegatableExpression)(node)) {
            const left = this.computeType(node.left);
            const right = this.computeType(node.right);
            if (left && right) {
                return this.computeTypeOfBinaryOperation(node.operator, left, right);
            }
            return undefined;
        }
        if ((0, ast_1.isBetweenExpression)(node)) {
            return sql_type_descriptors_1.Types.Boolean;
        }
        if ((0, ast_1.isSubQueryExpression)(node)) {
            return this.computeTypeOfSelectStatement(node.subQuery);
        }
        if ((0, ast_1.isIdentifierAsStringLiteral)(node)) {
            return sql_type_descriptors_1.Types.Char();
        }
        (0, langium_1.assertUnreachable)(node);
    }
    computeTypeOfSelectStatement(selectStatement) {
        return {
            discriminator: "row",
            columnTypes: (0, sql_type_utilities_1.getColumnsForSelectTableExpression)(selectStatement).map(c => ({
                name: c.name,
                type: this.computeType(c.typedNode)
            }))
        };
    }
    computeTypeOfDataType(dataType) {
        if (this.dataTypes.isBooleanDataType(dataType)) {
            return sql_type_descriptors_1.Types.Boolean;
        }
        if (this.dataTypes.isIntegerDataType(dataType)) {
            return sql_type_descriptors_1.Types.Integer;
        }
        if (this.dataTypes.isRealDataType(dataType)) {
            return sql_type_descriptors_1.Types.Real;
        }
        if (this.dataTypes.isStringDataType(dataType)) {
            return sql_type_descriptors_1.Types.Char();
        }
        if (this.dataTypes.isEnumDataType(dataType)) {
            return sql_type_descriptors_1.Types.Enum(dataType.arguments.map(e => e.value));
        }
        if (this.dataTypes.isDateTimeDataType(dataType)) {
            return sql_type_descriptors_1.Types.DateTime;
        }
        if (this.dataTypes.isBlobDataType(dataType)) {
            return sql_type_descriptors_1.Types.Blob;
        }
        return sql_type_descriptors_1.Types.Unknown;
    }
    computeTypeOfNumericLiteral(text) {
        var _a, _b, _c;
        const NumericLiteralPattern = /^(\d+)((\.(\d)+)?([eE]([\-+]?\d+))?)?$/;
        const match = NumericLiteralPattern.exec(text);
        const fractionalPart = (_b = (_a = match[4]) === null || _a === void 0 ? void 0 : _a.length) !== null && _b !== void 0 ? _b : 0;
        const exponent = parseInt((_c = match[6]) !== null && _c !== void 0 ? _c : "0", 10);
        return Math.max(0, fractionalPart - exponent) === 0 ? sql_type_descriptors_1.Types.Integer : sql_type_descriptors_1.Types.Real;
    }
    computeTypeOfBinaryOperation(operator, left, right) {
        const candidates = sql_type_operators_1.BinaryOperators[operator];
        for (const candidate of candidates) {
            if ((0, sql_type_descriptors_1.areTypesEqual)(candidate.left, left) && (0, sql_type_descriptors_1.areTypesEqual)(candidate.right, right)) {
                return candidate.returnType;
            }
            else {
                if ((0, sql_type_conversion_1.canConvert)(left, candidate.left, 'implicit') && (0, sql_type_conversion_1.canConvert)(right, candidate.right, 'implicit')) {
                    return candidate.returnType;
                }
            }
        }
        return undefined;
    }
    computeTypeOfUnaryOperation(operator, operandType) {
        const candidates = sql_type_operators_1.UnaryOperators[operator];
        for (const candidate of candidates) {
            if ((0, sql_type_descriptors_1.areTypesEqual)(candidate.operandType, operandType)) {
                return candidate.returnType;
            }
        }
        return undefined;
    }
}
exports.SqlTypeComputer = SqlTypeComputer;
//# sourceMappingURL=sql-type-computation.js.map