"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlValidator = exports.SqlValidationRegistry = void 0;
const tslib_1 = require("tslib");
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
const langium_1 = require("langium");
const ast = tslib_1.__importStar(require("./generated/ast"));
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const sql_error_codes_1 = require("./sql-error-codes");
const sql_type_descriptors_1 = require("./sql-type-descriptors");
const sql_type_utilities_1 = require("./sql-type-utilities");
const sql_type_conversion_1 = require("./sql-type-conversion");
const sql_data_types_1 = require("./sql-data-types");
class SqlValidationRegistry extends langium_1.ValidationRegistry {
    constructor(services) {
        super(services);
        const validator = services.validation.SqlValidator;
        const checks = {
            TableDefinition: [validator.checkIfTableDefinitionHasAtLeastOneColumn],
            SimpleSelectStatement: [
                validator.checkVariableNamesAreUnique,
                validator.checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource
            ],
            //Expression: [validator.checkIfExpressionHasType], //TODO uncomment when type system is bullet-proof
            BinaryTableExpression: [validator.checkBinaryTableExpressionMatches],
            NumberLiteral: [validator.checkIntegerLiteralIsWholeNumber],
            BinaryExpression: [validator.checkBinaryExpressionType],
            UnaryExpression: [validator.checkUnaryExpressionType],
            WhereClause: [validator.checkWhereIsBoolean],
            HavingClause: [validator.checkHavingIsBoolean],
            SubQueryExpression: [validator.checkIfSubQuerySelectsExactlyOneValue],
            ReferenceDefinition: [validator.checkIfReferencePointsToCorrectParent],
            FunctionCall: [validator.checkFunctionCallTarget],
            TableSourceItem: [validator.checkTableSourceItemTarget],
            ConstraintDefinition: [validator.checkConstraintTarget],
            DataType: [validator.checkIfKnownDataType],
        };
        this.register(checks, validator);
    }
}
exports.SqlValidationRegistry = SqlValidationRegistry;
class SqlValidator {
    constructor(services) {
        this.typeComputer = services.dialect.typeComputer;
        this.dataTypes = services.dialect.dataTypes.allTypes();
    }
    checkBinaryTableExpressionMatches(expr, accept) {
        const lhs = (0, sql_type_utilities_1.getColumnsForSelectTableExpression)(expr.left);
        const rhs = (0, sql_type_utilities_1.getColumnsForSelectTableExpression)(expr.right);
        if (lhs.length !== rhs.length) {
            sql_error_codes_1.ReportAs.TableOperationUsesTablesWithDifferentColumnCounts(expr, {}, accept);
        }
        else {
            lhs.forEach((left, index) => {
                const right = rhs[index];
                const leftType = this.typeComputer.computeType(left.typedNode);
                const rightType = this.typeComputer.computeType(right.typedNode);
                if (leftType && rightType) {
                    if (!(0, sql_type_conversion_1.canConvert)(leftType, rightType) && !(0, sql_type_conversion_1.canConvert)(rightType, leftType)) {
                        sql_error_codes_1.ReportAs.TableOperationUsesTablesWithDifferentColumnTypes(expr, { columnIndex: index }, accept);
                    }
                } //else should be handled by different validator
            });
        }
    }
    checkIfExpressionHasType(expr, accept) {
        const type = this.typeComputer.computeType(expr);
        if (!type) {
            sql_error_codes_1.ReportAs.CannotDeriveTypeOfExpression(expr, {}, accept);
        }
    }
    checkWhereIsBoolean(clause, accept) {
        const type = this.typeComputer.computeType(clause.rowCondition);
        if (type && !(0, sql_type_descriptors_1.isTypeABoolean)(type)) {
            sql_error_codes_1.ReportAs.ExpressionMustReturnABoolean(clause.rowCondition, type, accept);
        }
    }
    checkHavingIsBoolean(clause, accept) {
        const type = this.typeComputer.computeType(clause.groupCondition);
        if (type && !(0, sql_type_descriptors_1.isTypeABoolean)(type)) {
            sql_error_codes_1.ReportAs.ExpressionMustReturnABoolean(clause.groupCondition, type, accept);
        }
    }
    checkBinaryExpressionType(expr, accept) {
        const left = this.typeComputer.computeType(expr.left);
        const right = this.typeComputer.computeType(expr.right);
        const returnType = this.typeComputer.computeType(expr);
        if (left && right && !returnType) {
            sql_error_codes_1.ReportAs.BinaryOperatorNotDefinedForGivenExpressions(expr, {
                left,
                right,
                op: expr.operator
            }, accept);
        }
    }
    checkUnaryExpressionType(expr, accept) {
        const operand = this.typeComputer.computeType(expr.value);
        const returnType = this.typeComputer.computeType(expr);
        if (operand && !returnType) {
            sql_error_codes_1.ReportAs.UnaryOperatorNotDefinedForGivenExpression(expr, {
                operand,
                op: expr.operator
            }, accept);
        }
    }
    checkVariableNamesAreUnique(query, accept) {
        var _a, _b;
        const groups = lodash_1.default.groupBy((_b = (_a = query.from) === null || _a === void 0 ? void 0 : _a.sources) === null || _b === void 0 ? void 0 : _b.list, (s) => { var _a; return (_a = s.item) === null || _a === void 0 ? void 0 : _a.name; });
        for (const [key, group] of Object.entries(groups).filter((g) => g[0] && g[1].length > 1)) {
            for (const member of group.filter(e => e.item)) {
                sql_error_codes_1.ReportAs.DuplicatedVariableName(member.item, { name: key }, accept);
            }
        }
    }
    checkIntegerLiteralIsWholeNumber(literal, accept) {
        // if (Math.floor(literal.value) !== literal.value) {
        //     ReportAs.NumericValueIsNotInteger(
        //         literal,
        //         { value: literal.value },
        //         accept
        //     );
        // }
    }
    checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource(selectStatement, accept) {
        var _a;
        if (((_a = selectStatement.select) === null || _a === void 0 ? void 0 : _a.elements.filter(ast.isAllStar).length) > 0) {
            if (!selectStatement.from) {
                sql_error_codes_1.ReportAs.AllStarSelectionRequiresTableSources(selectStatement, {}, accept);
            }
        }
    }
    checkIfTableDefinitionHasAtLeastOneColumn(tableDefinition, accept) {
        if (tableDefinition.columns.length === 0) {
            sql_error_codes_1.ReportAs.TableDefinitionRequiresAtLeastOneColumn(tableDefinition.reference, {}, accept);
        }
    }
    checkIfReferencePointsToCorrectParent(referenceDefinition, accept) {
        this.checkReferenceParent(referenceDefinition.reference, accept);
    }
    checkReferenceParent(globalReference, accept) {
        var _a, _b, _c;
        const element = (_a = globalReference === null || globalReference === void 0 ? void 0 : globalReference.element) === null || _a === void 0 ? void 0 : _a.ref;
        if (element) {
            const parent = (_c = (_b = globalReference.previous) === null || _b === void 0 ? void 0 : _b.element) === null || _c === void 0 ? void 0 : _c.ref;
            if (parent) {
                const ownType = (0, sql_type_utilities_1.getDefinitionType)(element);
                const parentType = (0, sql_type_utilities_1.getDefinitionType)(parent);
                if (parentType >= ownType) {
                    accept('error', `Cannot nest element of type '${(0, sql_type_utilities_1.getDefinitionTypeName)(element)}' inside of '${(0, sql_type_utilities_1.getDefinitionTypeName)(parent)}'`, {
                        node: globalReference,
                        property: 'previous'
                    });
                }
                else if (globalReference.previous) {
                    this.checkReferenceParent(globalReference.previous, accept);
                }
            }
        }
    }
    //TODO does not hold for insertions! INSERT INTO employees SELECT id, name FROM xxx
    checkIfSubQuerySelectsExactlyOneValue(subQueryExpression, accept) {
        const type = this.typeComputer.computeTypeOfSelectStatement(subQueryExpression.subQuery);
        if (type.discriminator === 'row' && type.columnTypes.length > 1) {
            sql_error_codes_1.ReportAs.SubQueriesWithinSelectStatementsMustHaveExactlyOneColumn(subQueryExpression, {}, accept);
        }
    }
    checkFunctionCallTarget(functionCall, accept) {
        var _a, _b;
        const reference = (_b = (_a = functionCall.function) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isFunctionDefinition(reference)) {
            sql_error_codes_1.ReportAs.IncorrectGlobalReferenceTarget(functionCall.function, {
                expected: 'FUNCTION',
                received: (0, sql_type_utilities_1.getDefinitionTypeName)(reference)
            }, accept);
        }
    }
    checkTableSourceItemTarget(tableSourceItem, accept) {
        var _a, _b;
        const reference = (_b = (_a = tableSourceItem.table) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isTableLike(reference)) {
            sql_error_codes_1.ReportAs.IncorrectGlobalReferenceTarget(tableSourceItem.table, {
                expected: 'TABLE',
                received: (0, sql_type_utilities_1.getDefinitionTypeName)(reference)
            }, accept);
        }
    }
    checkConstraintTarget(constraintDefinition, accept) {
        var _a, _b;
        const reference = (_b = (_a = constraintDefinition.table) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isTableDefinition(reference)) {
            sql_error_codes_1.ReportAs.IncorrectGlobalReferenceTarget(constraintDefinition.table, {
                expected: 'TABLE',
                received: (0, sql_type_utilities_1.getDefinitionTypeName)(reference)
            }, accept);
        }
    }
    checkIfKnownDataType(dataType, accept) {
        if (!this.dataTypes.some(dt => (0, sql_data_types_1.isCompatibleWithDefinition)(dataType, dt))) {
            sql_error_codes_1.ReportAs.UnknownDataType(dataType, { dataType }, accept);
        }
    }
}
exports.SqlValidator = SqlValidator;
//# sourceMappingURL=sql-validator.js.map